"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getStringParameters = getStringParameters;
exports.getArgs = getArgs;
exports.getOptionsArg = getOptionsArg;
exports.getUserFuncArg = getUserFuncArg;
exports.getTypeDeclarationInBlock = getTypeDeclarationInBlock;
exports.getRegisterArguments = getRegisterArguments;
exports.getBlockParent = getBlockParent;
exports.getTypeParameter = getTypeParameter;
exports.throwMaybeAstError = exports.throwUnexpectedError = exports.Errors = exports.ErrorBase = void 0;

var _core = require("@babel/core");

var _babelPluginMacros = require("babel-plugin-macros");

var _commonTags = require("common-tags");

var _lodash = _interopRequireDefault(require("lodash.get"));

var _lodash2 = _interopRequireDefault(require("lodash.find"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 *  This file contains compile time helpers to validate that the macro is being called appropriately.
 */
// This is used in order to reduce duplication in the compile error tests
// If you update a message in here, the corresponding compile error test will pass automatically.
// No updating needed.
const ErrorBase = {
  ValidatorNoTypeParameter: `Failed to find type parameter. createValidator should be called with a type parameter. Example: "createValidator<TypeName>()"`,
  NotCalledAsFunction: `Macro should be called as function but was called as a`,
  MoreThanOneTypeParameter: `Macro should be called with 1 type parameter but was called with`,
  RegisterInvalidCall: (0, _commonTags.stripIndent)`
                                "register" should be called right below the type
                                declaration it is registering:
                                \`\`\`
                                interface Foo {
                                  bar: string
                                }
                                // no return value
                                register<Foo>()
                                \`\`\`
                                `,
  UnregisteredType: `Tried to generate a validator for an unregistered type with name`,
  // TODO: These will need to be updated once register accepts an options object
  RegisterInvalidNumberParams: `register should be called with 1 argument, but it was called with`,
  RegisterParam1NotStringLiteral: `register's first (and only) parameter should be a string literal, which is the name of the type to register, but it was a`,
  TypeDoesNotAcceptGenericParameters: `types don't accept generic parameters`,
  TooManyTypeParameters: `even though it only accepts`,
  NotEnoughTypeParameters: `type parameters even though it requires at least`,
  InvalidTypeParameterReference: `tried to reference the default type parameter in position:`
};
exports.ErrorBase = ErrorBase;

const removePeriod = str => str.slice(-1) === "." ? str.slice(0, -1) : str;

const addNewline = str => str.slice(-1) === "\n" ? str : str + "\n";

const Errors = {
  // Definitely user errors
  NoTypeParameters: () => ErrorBase.ValidatorNoTypeParameter,
  NotCalledAsFunction: callType => `${ErrorBase.NotCalledAsFunction} ${callType}`,
  MoreThanOneTypeParameter: typeParams => `${ErrorBase.MoreThanOneTypeParameter} ${typeParams}`,
  InvalidRegisterCall: () => ErrorBase.RegisterInvalidCall,
  // Errors that indicate possible bugs in the macro itself
  InvalidNumberOfRegisterParams: params => `${ErrorBase.RegisterInvalidNumberParams} ${params}`,
  RegisterParam1NotStringLiteral: paramNode => `${ErrorBase.RegisterParam1NotStringLiteral} ${paramNode.type}`,
  UnregisteredType: typeName => `${ErrorBase.UnregisteredType}: ${typeName}`,
  TypeDoesNotAcceptGenericParameters: (typeName, nodeType) => `Tried to instantiate "${typeName}" with generic parameters even though ${nodeType} ${ErrorBase.TypeDoesNotAcceptGenericParameters}`,
  TooManyTypeParameters: (typeName, provided, actual) => `Tried to instantiate ${typeName} with ${provided} type parameters ${ErrorBase.TooManyTypeParameters} ${actual}`,
  NotEnoughTypeParameters: (typeName, provided, actual) => `Tried to instantiate ${typeName} with ${provided} ${ErrorBase.NotEnoughTypeParameters} ${actual}`,
  InvalidTypeParameterReference: (paramPosition, referencedPosition) => (0, _commonTags.oneLine)`The default type parameter in position: ${paramPosition}
    ${ErrorBase.InvalidTypeParameterReference} ${referencedPosition}, which has not yet been instantiated`,
  UnexpectedError: reason => {
    return (0, _commonTags.oneLine)`Unexpected error because ${removePeriod(reason)}. Please report this to the developer.`;
  },
  // TODO: Can remove optional parameter (probably)
  MaybeAstError: (reason, optional) => {
    return (0, _commonTags.oneLine)`Invalid AST node because ${removePeriod(reason)}. Most likely, your Typescript is invalid.
    ${optional !== undefined ? addNewline(optional) : ""}But if it isn't, please report this to the developer.`;
  }
};
exports.Errors = Errors;

const throwUnexpectedError = msg => {
  throw new _babelPluginMacros.MacroError(Errors.UnexpectedError(msg));
};

exports.throwUnexpectedError = throwUnexpectedError;

const throwMaybeAstError = (msg, optional) => {
  throw new _babelPluginMacros.MacroError(Errors.MaybeAstError(msg, optional));
};

exports.throwMaybeAstError = throwMaybeAstError;

function assertCallExpr(expr) {
  if (!expr.isCallExpression()) throw new _babelPluginMacros.MacroError(Errors.NotCalledAsFunction(expr.type));
}

function getStringParameters(macroPath, functionName) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const args = callExpr.get("arguments");
  assertArray(args);
  const strings = [];

  for (const arg of args) {
    const {
      confident,
      value
    } = arg.evaluate();

    if (!confident || typeof value !== "string") {
      throwMaybeAstError(`the arguments to ${functionName} were not all string literals`);
    }

    strings.push(value);
  }

  return strings;
}

const transformerKeys = {
  c: "constraints"
};
// Keep this in sync with dist/typecheck.macro.d.ts
const keys = {
  c: "circularRefs",
  f: "allowForeignKeys",
  e: "expectedValueFormat"
};
const expectedValueFormats = ["human-friendly", "type-ir"];
const defaultOpts = {
  [keys.c]: true,
  [keys.f]: true,
  [keys.e]: expectedValueFormats[0]
};
Object.freeze(defaultOpts);

function getArgs(validatorType, macroPath, functionName, fileText, fileName) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const args = callExpr.get("arguments");
  assertArray(args);
  const defaultTransformers = {
    [transformerKeys.c]: new Map(),
    forbiddenVarNames: new Set()
  };

  if (args.length === 0) {
    return [defaultOpts, defaultTransformers];
  } else if (args.length === 1) {
    return [getOptionsArg(args[0], validatorType, functionName), defaultTransformers];
  } else if (args.length === 2) {
    return [getOptionsArg(args[0], validatorType, functionName), getUserFuncArg(args[1], functionName, fileText, fileName)];
  } else {
    throw new _babelPluginMacros.MacroError(`Recieved ${args.length} arguments, although ${functionName} accepts a max of 2 arguments.`);
  }
}

function getOptionsArg(path, validatorType, functionName) {
  const {
    confident,
    value
  } = path.evaluate();

  if (!confident) {
    throw new _babelPluginMacros.MacroError(`${functionName}'s first argument is 1. an options object that is statically known at compile time or 2. undefined`);
  }

  if (value === undefined) return defaultOpts;
  const opts = {
    [keys.f]: (0, _lodash.default)(value, keys.f, defaultOpts[keys.f]),
    [keys.e]: validatorType === "detailed" ? (0, _lodash.default)(value, keys.e, defaultOpts[keys.e]) : expectedValueFormats[0],
    [keys.c]: (0, _lodash.default)(value, keys.c, defaultOpts[keys.c])
  };

  const checkType = (type, key) => {
    if (typeof opts[key] !== type) {
      throw new _babelPluginMacros.MacroError(`key: ${key} in the options object must be undefined or a ${type}`);
    }
  };

  const booleanType = "boolean";
  checkType(booleanType, keys.f);
  checkType(booleanType, keys.c);
  checkType("string", keys.e);

  if (!expectedValueFormats.includes(opts[keys.e])) {
    throw new _babelPluginMacros.MacroError(`expected value format must be one of: ${JSON.stringify(expectedValueFormats)}`);
  }

  return opts;
}

function getUserFuncArg(path, functionName, fileText, filename) {
  if (!_core.types.isObjectExpression(path)) {
    throw new _babelPluginMacros.MacroError(`${functionName}'s second argument must be an object expression`);
  }

  const props = path.get("properties");
  assertArray(props);

  const getChild = key => {
    const prop = (0, _lodash2.default)(props, x => _core.types.isObjectProperty(x) && (0, _lodash.default)(x, "node.key.name", null) === key);

    if (prop !== undefined) {
      const value = prop.get("value");
      assertSingular(value);

      if (!_core.types.isObjectExpression(value)) {
        throw new _babelPluginMacros.MacroError(`key: "${key}" of ${functionName}'s second argument must be a object expression`);
      }

      return value.get("properties");
    }

    return [];
  };

  const constraintMap = new Map();
  const forbiddenVarNames = new Set();
  const transformers = {
    [transformerKeys.c]: constraintMap,
    forbiddenVarNames
  };
  const refiners = getChild(transformerKeys.c);

  for (const prop of refiners) {
    if (!_core.types.isObjectProperty(prop)) {
      throw new _babelPluginMacros.MacroError(`${functionName}'s second argument must be an object expression composed of key-value pairs, where the keys are statically known (not computed)`);
    }

    const typeName = (0, _lodash.default)(prop, "node.key.name", null);

    if (typeName === null) {
      throw new _babelPluginMacros.MacroError(`Failed to get key name when parsing 2nd parameter of ${functionName}`);
    }

    if (typeof typeName !== "string" && typeof typeName !== "number") {
      throw new _babelPluginMacros.MacroError(`Expected ${JSON.stringify(typeName)} to be string or number`);
    }

    const valuePath = prop.get("value");
    assertSingular(valuePath);
    const refinementValueError = new _babelPluginMacros.MacroError(`The values of the refinement object in the 2nd parameter of ${functionName} must be strings or function declarations`);

    if (_core.types.isArrowFunctionExpression(valuePath) || _core.types.isFunctionExpression(valuePath)) {
      const {
        start,
        end
      } = valuePath.node;

      if (start === null || end === null) {
        throw new _babelPluginMacros.MacroError(`Failed to extract function text from 2nd parameter of ${functionName}`);
      }

      valuePath.traverse({
        Identifier(path, _) {
          forbiddenVarNames.add(path.node.name);
        }

      });
      const functionText = "(" + fileText.slice(start, end) + ")"; // compile the function in order to remove type annotations

      const result = (0, _core.transformSync)(functionText, {
        filename
      });
      let code = result === null || result === void 0 ? void 0 : result.code;

      if (code === null || code === undefined) {
        throw new _babelPluginMacros.MacroError(`Failed to compile function pass as value of 2nd parameter of ${functionName}`);
      }

      code = code.replace(/^"use strict";\n*/, ""); // parenthesis will be kept around function expressions and not around arrow functions
      // arrow functions will be ended with a semi-colon

      if (code.slice(-1) === ";") code = code.slice(0, -1);
      code = "(" + code + ")";
      constraintMap.set(typeName, {
        value: code,
        type: "function"
      });
    } else {
      throw refinementValueError;
    }
  }

  return transformers;
}

function assertSingular(expr) {
  if (expr === null || expr === undefined || Array.isArray(expr)) {
    throw new _babelPluginMacros.MacroError(Errors.UnexpectedError(`expected expr to be single NodePath but it was ${expr}`));
  }
}

function assertArray(expr) {
  if (expr === null || expr === undefined || !Array.isArray(expr)) {
    throwUnexpectedError(`Expected expression to be array, but it was ${JSON.stringify(expr)}`);
  }
}

function getTypeDeclarationInBlock(typeName, stmts, idxInBlock) {
  for (let i = 0; i < stmts.length; ++i) {
    if (i === idxInBlock) continue;
    let stmt = stmts[i]; // If the statement is an export (named or default) use the declared statement instead

    if (_core.types.isExportNamedDeclaration(stmt) || _core.types.isExportDefaultDeclaration(stmt)) {
      stmt = stmt.declaration;
    }

    if (_core.types.isTSInterfaceDeclaration(stmt) || _core.types.isTSTypeAliasDeclaration(stmt)) {
      const declarationName = stmt.id.name;

      if (declarationName === typeName) {
        return stmt;
      }
    }
  }

  return null;
}

function getRegisterArguments(macroPath) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const args = callExpr.node.arguments;

  if (args.length != 1) {
    throw new _babelPluginMacros.MacroError(Errors.InvalidNumberOfRegisterParams(args.length));
  }

  const typeNamePath = callExpr.get("arguments.0");
  assertSingular(typeNamePath);

  if (!typeNamePath.isStringLiteral()) {
    throw new _babelPluginMacros.MacroError(Errors.RegisterParam1NotStringLiteral(typeNamePath.node));
  }

  const {
    confident,
    value: typeName
  } = typeNamePath.evaluate();

  if (confident === false || typeof typeName !== "string") {
    throwUnexpectedError(`evaluation result had type ${typeof typeName} and confidence ${confident}`);
  }

  return typeName;
}

function getBlockParent(macroPath) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const exprStmt = callExpr.parentPath;
  if (!exprStmt.isExpressionStatement()) throw new _babelPluginMacros.MacroError(Errors.InvalidRegisterCall());
  const {
    node
  } = exprStmt.parentPath;

  if (!_core.types.isProgram(node) && !_core.types.isBlock(node)) {
    throw new _babelPluginMacros.MacroError(Errors.InvalidRegisterCall());
  } else {
    return node.body;
  }
}

function getTypeParameter(macroPath) {
  const callExpr = macroPath.parentPath;
  assertCallExpr(callExpr);
  const typeParametersPath = callExpr.get("typeParameters");

  if (typeParametersPath && !Array.isArray(typeParametersPath) && typeParametersPath.node) {
    const {
      node
    } = typeParametersPath;

    if (_core.types.isTSTypeParameterInstantiation(node)) {
      const params = node.params.length;
      if (params != 1) throw new _babelPluginMacros.MacroError(Errors.MoreThanOneTypeParameter(params));
      const typeParameterPath = typeParametersPath.get("params.0");

      if (!Array.isArray(typeParameterPath) && // @ts-ignore: https://github.com/babel/babel/issues/11535
      typeParameterPath.isTSType()) {
        // @ts-ignore: https://github.com/babel/babel/issues/11535
        return typeParameterPath;
      }

      throwUnexpectedError(`typeParameter was ${Array.isArray(typeParameterPath) ? "an array" : "not a TSType"}`);
    }

    throwUnexpectedError(`typeParameters node was ${node.type} instead of TSTypeParameterInstantiation`);
  } else {
    throw new _babelPluginMacros.MacroError(Errors.NoTypeParameters());
  }
}